@@ -114,7 +114,7 @@ static int _handle_list_pids(int fd, stepd_step_rec_t *step);
 static int _handle_reconfig(int fd, stepd_step_rec_t *step, uid_t uid);
 static int _handle_get_ns_fd(int fd, stepd_step_rec_t *step);
 static bool _msg_socket_readable(eio_obj_t *obj);
-static int _msg_socket_accept(eio_obj_t *obj, List objs);
+static int _msg_socket_accept(eio_obj_t *obj, list_t *objs);
 
 struct io_operations msg_socket_ops = {
 	.readable = &_msg_socket_readable,
@@ -125,6 +125,11 @@ static char *socket_name;
 static pthread_mutex_t suspend_mutex = PTHREAD_MUTEX_INITIALIZER;
 static bool suspended = false;
 
+static int extern_thread_cnt = 0;
+static pthread_t *extern_threads = NULL;
+static pthread_mutex_t extern_thread_lock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_cond_t extern_thread_cond = PTHREAD_COND_INITIALIZER;
+
 struct request_params {
 	int fd;
 	stepd_step_rec_t *step;
@@ -374,8 +379,7 @@ _msg_socket_readable(eio_obj_t *obj)
 	return true;
 }
 
-static int
-_msg_socket_accept(eio_obj_t *obj, List objs)
+static int _msg_socket_accept(eio_obj_t *obj, list_t *objs)
 {
 	stepd_step_rec_t *step = obj->arg;
 	int fd;
@@ -502,7 +506,7 @@ static int _handle_stepmgr_relay_msg(int fd,
 	uint32_t data_size;
 
 	safe_read(fd, &protocol_version, sizeof(uint16_t));
-	client_fd = receive_fd_over_pipe(fd);
+	client_fd = receive_fd_over_socket(fd);
 	safe_read(fd, &data_size, sizeof(uint32_t));
 	data = xmalloc(data_size);
 	safe_read(fd, data, data_size);
@@ -560,12 +564,13 @@ static int _handle_step_create(int fd, stepd_step_rec_t *step, uid_t uid)
 	req_step_msg = msg.data;
 	slurm_mutex_lock(&stepmgr_mutex);
 	msg.auth_uid = req_step_msg->user_id = job_step_ptr->user_id;
-	slurm_mutex_unlock(&stepmgr_mutex);
 	msg.auth_ids_set = true;
 
 	/* step_create_from_msg responds to the client */
 	step_create_from_msg(&msg, NULL, NULL);
 
+	slurm_mutex_unlock(&stepmgr_mutex);
+
 	slurm_free_msg_members(&msg);
 
 	return SLURM_SUCCESS;
@@ -580,7 +585,6 @@ static int _handle_job_step_get_info(int fd, stepd_step_rec_t *step, uid_t uid)
 	buf_t *buffer;
 	slurm_msg_t msg;
 	job_step_info_request_msg_t *request;
-	slurm_msg_t response_msg;
 	pack_step_args_t args =  {0};
 
 	if ((rc = _handle_stepmgr_relay_msg(fd, uid, &msg,
@@ -602,12 +606,8 @@ static int _handle_job_step_get_info(int fd, stepd_step_rec_t *step, uid_t uid)
 	pack_job_step_info_response_msg(&args);
 	slurm_mutex_unlock(&stepmgr_mutex);
 
-	response_init(&response_msg, &msg, RESPONSE_JOB_STEP_INFO,
-		      buffer);
-	slurm_send_node_msg(msg.conn_fd, &response_msg);
+	(void) send_msg_response(&msg, RESPONSE_JOB_STEP_INFO, buffer);
 	FREE_NULL_BUFFER(buffer);
-
-	slurm_send_rc_msg(&msg, SLURM_SUCCESS);
 	slurm_free_msg_members(&msg);
 
 done:
@@ -746,14 +746,13 @@ static int _handle_step_layout(int fd, stepd_step_rec_t *step, uid_t uid)
 	rc = stepmgr_get_step_layouts(job_step_ptr, request, &step_layout);
 	slurm_mutex_unlock(&stepmgr_mutex);
 	if (!rc) {
-		slurm_msg_t response_msg;
-		response_init(&response_msg, &msg, RESPONSE_STEP_LAYOUT,
-			      step_layout);
-		slurm_send_node_msg(msg.conn_fd, &response_msg);
+		(void) send_msg_response(&msg, RESPONSE_STEP_LAYOUT,
+					 step_layout);
 		slurm_step_layout_destroy(step_layout);
+	} else {
+		slurm_send_rc_msg(&msg, rc);
 	}
 
-	slurm_send_rc_msg(&msg, rc);
 	slurm_free_msg_members(&msg);
 
 done:
@@ -766,7 +765,6 @@ static int _handle_job_sbcast_cred(int fd, stepd_step_rec_t *step, uid_t uid)
 	slurm_msg_t msg;
 	step_alloc_info_msg_t *request;
 	job_sbcast_cred_msg_t *job_info_resp_msg = NULL;
-	slurm_msg_t response_msg;
 
 	if ((rc = _handle_stepmgr_relay_msg(fd, uid, &msg,
 					    REQUEST_JOB_SBCAST_CRED, true)))
@@ -782,12 +780,12 @@ static int _handle_job_sbcast_cred(int fd, stepd_step_rec_t *step, uid_t uid)
 	if (rc)
 		goto resp;
 
-	response_init(&response_msg, &msg, RESPONSE_JOB_SBCAST_CRED,
-		      job_info_resp_msg);
-
-	slurm_send_node_msg(msg.conn_fd, &response_msg);
+	(void) send_msg_response(&msg, RESPONSE_JOB_SBCAST_CRED,
+				 job_info_resp_msg);
 
 	slurm_free_sbcast_cred_msg(job_info_resp_msg);
+	slurm_free_msg_members(&msg);
+	return rc;
 
 resp:
 	slurm_send_rc_msg(&msg, rc);
@@ -809,8 +807,7 @@ static int _handle_het_job_alloc_info(int fd, stepd_step_rec_t *step, uid_t uid)
 	slurm_msg_t msg;
 	job_alloc_info_msg_t *request;
 	resource_allocation_response_msg_t *job_info_resp_msg = NULL;
-	slurm_msg_t response_msg;
-	List resp_list;
+	list_t *resp_list = NULL;
 
 	if ((rc = _handle_stepmgr_relay_msg(fd, uid, &msg,
 					    REQUEST_HET_JOB_ALLOC_INFO, true)))
@@ -834,10 +831,11 @@ static int _handle_het_job_alloc_info(int fd, stepd_step_rec_t *step, uid_t uid)
 
 	slurm_mutex_unlock(&stepmgr_mutex);
 
-	response_init(&response_msg, &msg, RESPONSE_HET_JOB_ALLOCATION,
-		      resp_list);
-	slurm_send_node_msg(msg.conn_fd, &response_msg);
+	(void) send_msg_response(&msg, RESPONSE_HET_JOB_ALLOCATION, resp_list);
+
 	FREE_NULL_LIST(resp_list);
+	slurm_free_msg_members(&msg);
+	return rc;
 
 resp:
 	slurm_send_rc_msg(&msg, rc);
@@ -976,7 +974,7 @@ int _handle_request(int fd, stepd_step_rec_t *step, uid_t uid, pid_t remote_pid)
 		_handle_srun_timeout(fd, step, uid);
 		break;
 	case REQUEST_UPDATE_JOB_STEP:
-		debug("Handling REQUEST_CANCEL_JOB_STEP");
+		debug("Handling REQUEST_UPDATE_JOB_STEP");
 		rc = _handle_update_step(fd, step, uid);
 		break;
 	case REQUEST_STEP_LAYOUT:
@@ -1495,12 +1493,12 @@ static int _handle_get_ns_fd(int fd, stepd_step_rec_t *step)
 
 	/*
 	 * We need to send the ns_fd as an int first to let the receiver know if
-	 * we have a valid fd or not as recv_fd_over_pipe() will always try to
-	 * set up the fd no matter if it is valid or not.
+	 * we have a valid fd or not as receive_fd_over_socket() will always
+	 * try to set up the fd no matter if it is valid or not.
 	 */
 	safe_write(fd, &ns_fd, sizeof(ns_fd));
 	if (ns_fd > 0)
-		send_fd_over_pipe(fd, ns_fd);
+		send_fd_over_socket(fd, ns_fd);
 
 	debug("sent fd: %d", ns_fd);
 	debug("leaving %s", __func__);
@@ -1510,13 +1508,20 @@ rwfail:
 	return SLURM_ERROR;
 }
 
-static void _block_on_pid(pid_t pid)
+static void _block_on_pid(pid_t pid, stepd_step_rec_t *step)
 {
-	/* I wish there was another way to wait on a foreign pid, but
-	 * I was unable to find one.
-	 */
-	while (kill(pid, 0) != -1)
-		sleep(1);
+	struct timespec ts = { 0, 0 };
+
+	slurm_mutex_lock(&extern_thread_lock);
+	while (kill(pid, 0) != -1) {
+		if (step->state >= SLURMSTEPD_STEP_CANCELLED)
+			break;
+		clock_gettime(CLOCK_REALTIME, &ts);
+		ts.tv_sec += 1;
+		slurm_cond_timedwait(&extern_thread_cond, &extern_thread_lock,
+				     &ts);
+	}
+	slurm_mutex_unlock(&extern_thread_lock);
 }
 
 /* Wait for the pid given and when it ends get and children it might
@@ -1541,7 +1546,7 @@ static void *_wait_extern_pid(void *args)
 	xfree(extern_pid);
 
 	//info("waiting on pid %d", pid);
-	_block_on_pid(pid);
+	_block_on_pid(pid, step);
 	//info("done with pid %d %d: %m", pid, rc);
 	jobacct = jobacct_gather_remove_task(pid);
 	if (jobacct) {
@@ -1551,6 +1556,9 @@ static void *_wait_extern_pid(void *args)
 	}
 	acct_gather_profile_g_task_end(pid);
 
+	if (step->state >= SLURMSTEPD_STEP_CANCELLED)
+		goto end;
+
 	/* See if we have any children of init left and add them to track. */
 	proctrack_g_get_pids(step->cont_id, &pids, &npids);
 	for (i = 0; i < npids; i++) {
@@ -1584,11 +1592,22 @@ static void *_wait_extern_pid(void *args)
 	next_pid:
 		fclose(stat_fp);
 	}
+end:
 	xfree(pids);
-
 	return NULL;
 }
 
+static void _wait_extern_thr_create(extern_pid_t *extern_pid)
+{
+	/* Lock as several RPC can write to the same variable. */
+	slurm_mutex_lock(&extern_thread_lock);
+	extern_thread_cnt++;
+	xrecalloc(extern_threads, extern_thread_cnt, sizeof(pthread_t));
+	slurm_thread_create(&extern_threads[extern_thread_cnt - 1],
+			    _wait_extern_pid, extern_pid);
+	slurm_mutex_unlock(&extern_thread_lock);
+}
+
 static int _handle_add_extern_pid_internal(stepd_step_rec_t *step, pid_t pid)
 {
 	extern_pid_t *extern_pid;
@@ -1635,7 +1654,7 @@ static int _handle_add_extern_pid_internal(stepd_step_rec_t *step, pid_t pid)
 		set_user_limits(step, pid);
 
 	/* spawn a thread that will wait on the pid given */
-	slurm_thread_create_detached(_wait_extern_pid, extern_pid);
+	_wait_extern_thr_create(extern_pid);
 
 	return SLURM_SUCCESS;
 }
@@ -1645,6 +1664,13 @@ static int _handle_add_extern_pid(int fd, stepd_step_rec_t *step, uid_t uid)
 	int rc = SLURM_SUCCESS;
 	pid_t pid;
 
+	slurm_mutex_lock(&step->state_mutex);
+	if (step->state >= SLURMSTEPD_STEP_CANCELLED) {
+		error("Rejecting request to add extern pid from uid %u because step is ending",
+		      uid);
+		goto rwfail;
+	}
+
 	safe_read(fd, &pid, sizeof(pid_t));
 
 	if (!_slurm_authorized_user(uid)) {
@@ -1658,8 +1684,10 @@ static int _handle_add_extern_pid(int fd, stepd_step_rec_t *step, uid_t uid)
 	safe_write(fd, &rc, sizeof(int));
 
 	debug("Leaving _handle_add_extern_pid");
+	slurm_mutex_unlock(&step->state_mutex);
 	return SLURM_SUCCESS;
 rwfail:
+	slurm_mutex_unlock(&step->state_mutex);
 	return SLURM_ERROR;
 }
 
@@ -1803,7 +1831,7 @@ static int _handle_getgr(int fd, stepd_step_rec_t *step, pid_t remote_pid)
 
 	if (!step->ngids || !step->gids || !step->gr_names) {
 		error("%s: incomplete data, ignoring request", __func__);
-	} else if (mode == GETGR_MATCH_GROUP_AND_PID ) {
+	} else if ((mode == GETGR_MATCH_GROUP_AND_PID) && pid_match) {
 		while (offset < step->ngids) {
 			if (gid == step->gids[offset])
 				break;
@@ -2162,31 +2190,39 @@ _handle_completion(int fd, stepd_step_rec_t *step, uid_t uid)
 		goto rwfail;
 	FREE_NULL_BUFFER(buffer);
 
-	if (job_step_ptr && do_stepmgr) {
-		int rem = 0;
-		uint32_t max_rc;
-		slurm_step_id_t temp_id = {
-			.job_id = job_step_ptr->job_id,
-			.step_het_comp = NO_VAL,
-			.step_id = step_id
-		};
+	if (do_stepmgr) {
+		slurm_mutex_lock(&stepmgr_mutex);
+		if (job_step_ptr) {
+			int rem = 0;
+			uint32_t max_rc;
+			slurm_step_id_t temp_id = {
+				.job_id = job_step_ptr->job_id,
+				.step_het_comp = NO_VAL,
+				.step_id = step_id
+			};
 
-		step_complete_msg_t req = {
-			.range_first = first,
-			.range_last = last,
-			.step_id = temp_id,
-			.step_rc = step_rc,
-			.jobacct = jobacct
-		};
+			step_complete_msg_t req = {
+				.range_first = first,
+				.range_last = last,
+				.step_id = temp_id,
+				.step_rc = step_rc,
+				.jobacct = jobacct
+			};
 
-		step_partial_comp(&req, uid, true, &rem, &max_rc);
+			step_partial_comp(&req, uid, true, &rem, &max_rc);
 
-		safe_write(fd, &rc, sizeof(int));
-		safe_write(fd, &errnum, sizeof(int));
+			safe_write(fd, &rc, sizeof(int));
+			safe_write(fd, &errnum, sizeof(int));
 
-		jobacctinfo_destroy(jobacct);
+			jobacctinfo_destroy(jobacct);
 
-		return SLURM_SUCCESS;
+			rc = SLURM_SUCCESS;
+		} else {
+			error("Asked to complete a stepmgr step but we don't have a job_step_ptr. This should never happen.");
+			rc = SLURM_ERROR;
+		}
+		slurm_mutex_unlock(&stepmgr_mutex);
+		return rc;
 	}
 
 	/*
@@ -2477,3 +2513,25 @@ extern void set_msg_node_id(stepd_step_rec_t *step)
 	if (ptr)
 		msg_target_node_id = atoi(ptr);
 }
+
+extern void join_extern_threads()
+{
+	int th_cnt;
+
+	slurm_mutex_lock(&extern_thread_lock);
+	slurm_cond_broadcast(&extern_thread_cond);
+	th_cnt = extern_thread_cnt;
+	slurm_mutex_unlock(&extern_thread_lock);
+
+	for (int i = 0; i < th_cnt; i++) {
+		debug2("Joining extern pid thread %d", i);
+		slurm_thread_join(extern_threads[i]);
+	}
+
+	slurm_mutex_lock(&extern_thread_lock);
+	extern_thread_cnt = 0;
+	xfree(extern_threads);
+	slurm_mutex_unlock(&extern_thread_lock);
+
+	debug2("Done joining extern pid threads");
+}
